/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2020 Chris Holland <zrenfire@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

import QtQuick 2.0
import QtQuick.Layouts 1.1
import org.kde.plasma.core 2.0 as PlasmaCore
import org.kde.plasma.components 2.0 as PlasmaComponents
import org.kde.kquickcontrolsaddons 2.0
import org.ukui.kwin 2.0 as KWin
import QtGraphicalEffects 1.0
import QtQuick.Controls 2.0

// https://techbase.kde.org/Development/Tutorials/KWin/WindowSwitcher
// https://github.com/KDE/kwin/blob/master/tabbox/switcheritem.h


KWin.Switcher {
    id: tabBox
    currentIndex: thumbnailGridView.currentIndex
    
    

    PlasmaCore.Dialog {
        id: dialog
        location: PlasmaCore.Types.Floating
        visible: tabBox.visible
        backgroundHints: PlasmaCore.Dialog.NoBackground
        flags: Qt.X11BypassWindowManagerHint
        x: tabBox.unifiedGeometry.x
        y: tabBox.unifiedGeometry.y

        onVisibleChanged: {
            if (visible) {
                dialogMainItem.calculateColumnCount();
            } else {
                thumbnailGridView.highCount = 0;
            }
        }    

        
        mainItem: Rectangle {
            id: dialogMainItem
            focus: true
            Keys.enabled: true
            color: Qt.rgba(35/255, 36/255, 38/255, 0.45)

            MouseArea {
                anchors.fill:parent;
                    onClicked: {
                        tabBox.model.hide();
                    }
                }

            property int maxWidth: tabBox.unifiedGeometry.width //1 控制窗口位置
            property int maxHeight: tabBox.unifiedGeometry.height * 0.8 // 1
            property real screenFactor: tabBox.unifiedGeometry.width / tabBox.unifiedGeometry.height
            property int maxGridColumnsByWidth: Math.floor(maxWidth / thumbnailGridView.cellWidth)

            property int gridColumns: maxGridColumnsByWidth
            property int gridRows: Math.ceil(thumbnailGridView.count / gridColumns)
            property int optimalWidth: thumbnailGridView.cellWidth * gridColumns
            property int optimalHeight: thumbnailGridView.cellHeight * gridRows
            property bool canStretchX: false
            property bool canStretchY: false
            width : tabBox.unifiedGeometry.width
            height : tabBox.unifiedGeometry.height
            
            clip: true


            // Simple greedy algorithm
            function calculateColumnCount() {
//                // respect screenGeometry
//                const c = Math.min(thumbnailGridView.count, maxGridColumnsByWidth);

//                const residue = thumbnailGridView.count % c;
//                if (residue == 0) {
//                    gridColumns = c;
//                    return;
//                }

//                // start greedy recursion
//                gridColumns = columnCountRecursion(c, c, c - residue);
            }

            // Step for greedy algorithm
            function columnCountRecursion(prevC, prevBestC, prevDiff) {
                const c = prevC - 1;

                // don't increase vertical extent more than horizontal
                // and don't exceed maxHeight
                if (prevC * prevC <= thumbnailGridView.count + prevDiff ||
                        maxHeight < Math.ceil(thumbnailGridView.count / c) * thumbnailGridView.cellHeight) {
                    return prevBestC;
                }
                const residue = thumbnailGridView.count % c;
                // halts algorithm at some point
                if (residue == 0) {
                    return c;
                }
                // empty slots
                const diff = c - residue;

                // compare it to previous count of empty slots
                if (diff < prevDiff) {
                    return columnCountRecursion(c, c, diff);
                } else if (diff == prevDiff) {
                    // when it's the same try again, we'll stop early enough thanks to the landscape mode condition
                    return columnCountRecursion(c, prevBestC, diff);
                }
                // when we've found a local minimum choose this one (greedy)
                return columnCountRecursion(c, prevBestC, diff);
            }

            // Just to get the margin sizes
//            PlasmaCore.FrameSvgItem {
//                id: hoverItem
//                imagePath: "widgets/viewitem"
//                prefix: "hover"
//                visible: false


//            }

            GridView {
                id: thumbnailGridView
                anchors.fill: parent

                anchors.bottomMargin: 10 * units.devicePixelRatio
                anchors.topMargin: 10 * units.devicePixelRatio
                anchors.leftMargin: count > dialogMainItem.maxGridColumnsByWidth ? minLeftMargin : (dialogMainItem.width-count * cellWidth)/2
                anchors.rightMargin: count > dialogMainItem.maxGridColumnsByWidth ? parent.right : 1

                model: tabBox.model


                MouseArea {
                        anchors.fill: parent
                        propagateComposedEvents: true
                        onClicked: {
                            mouse.accepted=false
                        }
                }


                property int iconSize: 24 * units.devicePixelRatio
                property int captionRowHeight: 32 * units.devicePixelRatio // The close button is 30x30 in Breeze
                property int thumbnailWidth: (tabBox.screenGeometry.width-160)/4
               // property int thumbnailHeight: thumbnailWidth * (1.0/dialogMainItem.screenFactor)
                property int thumbnailHeight: tabBox.screenGeometry.height/22*5
                cellWidth: 8 * units.devicePixelRatio + thumbnailWidth + 16 * units.devicePixelRatio
                cellHeight: 8 * units.devicePixelRatio + captionRowHeight + thumbnailHeight + 8 * units.devicePixelRatio
                property int  minLeftMargin : tabBox.unifiedGeometry.width > thumbnailWidth  ? (dialogMainItem.width-dialogMainItem.maxGridColumnsByWidth * cellWidth)/2 : 1


                keyNavigationWraps: true
                highlightMoveDuration: 0

                // Allow expansion on increasing count
                property int highCount: 0
                onCountChanged: {
                    if (highCount < count) {
                        dialogMainItem.calculateColumnCount();
                        highCount = count;
                    }
                    if (count <= dialogMainItem.maxGridColumnsByWidth) {

                    }

                }

                ScrollBar.vertical: ScrollBar {
                    onActiveChanged: {
                        active = true;
                    }
                    policy: ScrollBar.AlwaysOn      // 滚动条ce'l策略，这里是永远显示
                    // policy: ScrollBar.AsNeeded  环境变量会导致初始化时无法显示，激活后才能正常
                }

                delegate: Item {
                    id: thumbnailGridItem
                    width: thumbnailGridView.cellWidth
                    height: thumbnailGridView.cellHeight

                    MouseArea {
                        anchors.fill: parent
                        // propagateComposedEvents: true //鼠标事件传递
                        onClicked: {
                            thumbnailGridItem.select();
                            tabBox.model.choose();
                            // mouse.accepted=false
                        }
                        hoverEnabled: true

                        onEntered: {
                            thumbnailBorderItem.border.width = 4;

                        }

                        onExited: {
                            thumbnailBorderItem.border.width = 0;

                        }
                    }

                    function select() {
                        thumbnailGridView.currentIndex = index;
                        thumbnailGridView.currentIndexChanged(thumbnailGridView.currentIndex);
                    }

                    function getThumbnailSize()
                    {
                        var thumbHeight = thumbnailGridView.thumbnailHeight
                        var max_width = thumbnailGridView.thumbnailWidth
                        var clientWidth = model.width
                        var clientHeight = model.height
                        var thumbWidth = Math.floor(thumbHeight * clientWidth / clientHeight)
                        if (thumbWidth > max_width) {
                            thumbWidth = max_width
                            thumbHeight = Math.floor(thumbWidth * clientHeight / clientWidth)
                        }
                        return [thumbWidth, thumbHeight]
                    }


                    ColumnLayout {
                        z: 0
                        spacing: 0
                        anchors.fill: parent
                        anchors.leftMargin: 8 * units.devicePixelRatio
                        anchors.topMargin: 8 * units.devicePixelRatio
                        anchors.rightMargin: 8 * units.devicePixelRatio
                        anchors.bottomMargin: 8 * units.devicePixelRatio

                        RowLayout {
                            id: captionRow
                            spacing: 4 * units.devicePixelRatio

                            QIconItem {
                                id: iconItem
                                Layout.minimumHeight: thumbnailGridView.iconSize
                                Layout.minimumWidth: thumbnailGridView.iconSize
                                Layout.maximumHeight: Layout.minimumHeight
                                Layout.maximumWidth: Layout.minimumWidth
                                icon: model.icon
                                state: index == thumbnailGridView.currentIndex ? QIconItem.ActiveState : QIconItem.DefaultState
                            }

                            PlasmaComponents.Label {
                                id: label
                                Layout.fillWidth: true
                                height: 20
                                color: "#FFFFFF"
                                text: model.caption
                                elide: Text.ElideRight
                            }

                            PlasmaComponents.ToolButton {
                                id: closeButton
                                visible: false
                                iconSource: 'window-close-symbolic'
                                onClicked: {
                                    tabBox.model.close(index);
                                }

                            }
                        }

                        // KWin.ThumbnailItem needs a container
                        // otherwise it will be drawn the same size as the parent ColumnLayout
                        Rectangle {
                            id: thumbnailBorderItem
                            Layout.fillWidth: true
                            Layout.fillHeight: true
                            color: "transparent" //小框背景
                            border.color: "#FFFFFF";
                            border.width: 0;
                            radius: 8;
                            Rectangle {
                                id: thumbnailBackItem
                                width: getThumbnailSize()[0]
                                height: getThumbnailSize()[1]
                                anchors.centerIn: parent
                                color: "transparent"
                                clip: true

                                layer.enabled: tabBox.isOpenGL ? true : false
                                layer.effect: OpacityMask {
                                maskSource: Rectangle {
                                        width: getThumbnailSize()[0]
                                        height: getThumbnailSize()[1]
                                        radius: 8
                                    }
                                }

                                // Cannot draw anything (like an icon) on top of thumbnail
                                KWin.ThumbnailItem {
                                    id: thumbnailItem
                                    anchors.centerIn: parent
                                    width: getThumbnailSize()[0]
                                    height: getThumbnailSize()[1]
                                    wId: windowId
                                    clip: true
                                }
                            }
                        }

                    }
                } // GridView.delegate

                highlight: Rectangle {
                    id: highlightItem
                    color: "transparent"
                    visible : tabBox.highlightVisible
                    border.color: "#FFFFFF"
                    border.width:4;
                    radius: 8;

                }
                clip:true

                Connections {
                    target: tabBox
                    onCurrentIndexChanged: {
                        thumbnailGridView.currentIndex = tabBox.currentIndex;
                    }
                }
            } // GridView


            Keys.onPressed: {
                if (event.key == Qt.Key_Left) {
                    thumbnailGridView.moveCurrentIndexLeft();
                } else if (event.key == Qt.Key_Right) {
                    thumbnailGridView.moveCurrentIndexRight();
                } else if (event.key == Qt.Key_Up) {
                    thumbnailGridView.moveCurrentIndexUp();
                } else if (event.key == Qt.Key_Down) {
                    thumbnailGridView.moveCurrentIndexDown();
                } else if (event.key == Qt.Key_Escape || event.key == Qt.Key_Super_L || event.key == Qt.Key_Super_R) {
                    tabBox.model.hide();
                } else if (event.key == Qt.Key_Enter || event.key == Qt.Key_Return || event.key == Qt.Key_Space) {
                    tabBox.model.choose();
                } else {
                    return;
                }
                thumbnailGridView.currentIndexChanged(thumbnailGridView.currentIndex);
            }
        } // Dialog.mainItem
    } // Dialog 
}
